@@ -30,7 +30,7 @@ def default(val, d):
     key=['M', 'N', 'K'],
 )
 @triton.jit
-def fused_relu_squared_kernel_forward(
+def bmm_kernel(
     x_ptr, y_ptr, o_ptr,
     M, N, K,
     stride_al, stride_am, stride_ak,
@@ -61,16 +61,17 @@ def fused_relu_squared_kernel_forward(
     x_ptrs = x_ptr + (offs_am[:, None]*stride_am + offs_k [None, :]*stride_ak + pid_batch*stride_al)
     y_ptrs = y_ptr + (offs_k [:, None]*stride_bk + offs_bn[None, :]*stride_bn + pid_batch*stride_bl)
 
-    c = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=tl.float32)
+    o = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=tl.float32)
     for k in range(0, K, BLOCK_SIZE_K):
         x = tl.load(x_ptrs)
         y = tl.load(y_ptrs)
-        c += tl.dot(x, y)
+        o += tl.dot(x, y)
 
         x_ptrs += BLOCK_SIZE_K * stride_ak
         y_ptrs += BLOCK_SIZE_K * stride_bk
 
-    o = tl.where(c > 0, c * c, 0.) # ReLU squared
+    if exists(meta['ACTIVATION']):
+        o = meta['ACTIVATION'](o)
 
     offs_m = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
     offs_n = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
@@ -79,33 +80,39 @@ def fused_relu_squared_kernel_forward(
     o_ptrs = o_ptr + stride_om * offs_m[:, None] + stride_on * offs_n[None, :] + stride_ol * pid_batch
     tl.store(o_ptrs, o, mask=mask)
 
+@triton.jit
+def relu_squared_activation(x):
+    return tl.where(x > 0, x * x, 0.)
+
+def triton_bmm(x, y, activation = None):
+    B, M, K = x.shape
+
+    if y.ndim == 2:
+        y = y.unsqueeze(0).expand(B, -1, -1)
+
+    _, K, N = y.shape
+    assert (K % 32 == 0), "K must be divisible by 32"
+
+    o = torch.empty((B, M, N), device = x.device, dtype = x.dtype)
+
+    grid = lambda META: (
+        B, triton.cdiv(M, META['BLOCK_SIZE_M']) * triton.cdiv(N, META['BLOCK_SIZE_N']),
+    )
+
+    bmm_kernel[grid](
+        x, y, o,
+        M, N, K,
+        x.stride(0), x.stride(1), x.stride(2),
+        y.stride(0), y.stride(1), y.stride(2),
+        o.stride(0), o.stride(1), o.stride(2),
+        ACTIVATION = activation
+    )
+    return o
+
 class _relu_squared(autograd.Function):
     @classmethod
     def forward(self, ctx, x, w):
-        B, N, D = x.shape
-
-        y = w[None, ...].expand(B, -1, -1)
-
-        B, M, K = x.shape
-        _, K, N = y.shape
-
-        assert (K % 32 == 0), "We don't check memory-out-of-bounds with K so K must be divisible by BLOCK_SIZE_K"
-
-        c = torch.empty((B, M, N), device = x.device, dtype = x.dtype)
-        o = torch.empty((B, M, N), device = x.device, dtype = x.dtype)
-
-        grid = lambda META: (
-            B, triton.cdiv(M, META['BLOCK_SIZE_M']) * triton.cdiv(N, META['BLOCK_SIZE_N']),
-        )
-
-        fused_relu_squared_kernel_forward[grid](
-            x, y, o,
-            M, N, K,
-            x.stride(0), x.stride(1), x.stride(2),
-            y.stride(0), y.stride(1), y.stride(2),
-            o.stride(0), o.stride(1), o.stride(2)
-        )
-
+        o = triton_bmm(x, w, activation = relu_squared_activation)
         ctx.save_for_backward(x, w, o)
         return o
 
@@ -113,8 +120,8 @@ class _relu_squared(autograd.Function):
     def backward(self, ctx, dy):
         x, w, o = ctx.saved_tensors
         dy = torch.sqrt(o) * 2 * dy
-        dx = dy @ w.t()
-        dw = x.transpose(-1, -2) @ dy
+        dx = triton_bmm(dy, w.t())
+        dw = triton_bmm(x.transpose(-1, -2), dy)
         return dx, dw
 
 triton_relu_squared = _relu_squared.apply
@@ -273,7 +280,7 @@ def cross_entropy_fn(logits, labels, ignore_index = 0., use_triton = False):
 # triton - layer norm
 
 @triton.jit
-def layernorm_kernel_forward(
+def layernorm_kernel_forward_training(
     output_ptr,
     normed_ptr,
     mean_centered_ptr,
@@ -333,6 +340,49 @@ def layernorm_kernel_forward(
     inv_var_ptrs = inv_var_row_start_ptr + col_offsets
     tl.store(inv_var_ptrs, inv_var, mask=mask)
 
+@triton.jit
+def layernorm_kernel_forward_inference(
+    output_ptr,
+    input_ptr,
+    gamma_ptr,
+    beta_ptr,
+    input_row_stride,
+    gamma_row_stride,
+    beta_row_stride,
+    output_row_stride,
+    n_cols,
+    eps,
+    **meta
+):
+    row_idx = tl.program_id(0)
+    BLOCK_SIZE = meta['BLOCK_SIZE']
+
+    row_start_ptr = input_ptr + row_idx * input_row_stride
+    gamma_row_start_ptr = gamma_ptr + row_idx * gamma_row_stride
+    beta_row_start_ptr = beta_ptr + row_idx * beta_row_stride
+
+    col_offsets = tl.arange(0, BLOCK_SIZE)
+    input_ptrs = row_start_ptr + col_offsets
+    gamma_ptrs = gamma_row_start_ptr + col_offsets
+    beta_ptrs = beta_row_start_ptr + col_offsets
+
+    mask = col_offsets < n_cols
+    row = tl.load(input_ptrs, mask=mask, other=0.)
+    gammas = tl.load(gamma_ptrs, mask=mask, other=0.)
+    betas = tl.load(beta_ptrs, mask=mask, other=0.)
+
+    row_mean = tl.sum(row, axis = 0) / n_cols
+    row_mean_centered = row - row_mean
+    row_var = tl.sum(row_mean_centered * row_mean_centered, axis = 0) / n_cols
+    inv_var = 1. / tl.sqrt(row_var + eps)
+    normed = row_mean_centered * inv_var
+
+    output = normed * gammas + betas
+
+    output_row_start_ptr = output_ptr + row_idx * output_row_stride
+    output_ptrs = output_row_start_ptr + col_offsets
+    tl.store(output_ptrs, output, mask=mask)
+
 @triton.jit
 def layernorm_kernel_backward(
     output_ptr,
@@ -377,7 +427,7 @@ def layernorm_kernel_backward(
 
 class _layernorm(autograd.Function):
     @classmethod
-    def forward(cls, ctx, x, gamma, beta, eps):
+    def forward(cls, ctx, x, gamma, beta, eps, training):
         shape = x.shape
         dim = shape[-1]
         x = x.view(-1, dim)
@@ -390,32 +440,49 @@ class _layernorm(autograd.Function):
         num_warps = calc_num_warps(BLOCK_SIZE)
 
         out = torch.empty_like(x)
-        normed_x = torch.empty_like(x)
-        scaled_x = torch.empty_like(x)
-        inv_var = torch.empty_like(x)
 
-        layernorm_kernel_forward[(n_rows,)](
-            out,
-            normed_x,
-            scaled_x,
-            inv_var,
-            x,
-            expanded_gamma,
-            expanded_beta,
-            x.stride(0),
-            expanded_gamma.stride(0),
-            expanded_beta.stride(0),
-            out.stride(0),
-            normed_x.stride(0),
-            scaled_x.stride(0),
-            inv_var.stride(0),
-            n_cols,
-            eps,
-            num_warps = num_warps,
-            BLOCK_SIZE = BLOCK_SIZE,
-        )
+        if training:
+            normed_x = torch.empty_like(x)
+            scaled_x = torch.empty_like(x)
+            inv_var = torch.empty_like(x)
+
+            layernorm_kernel_forward_training[(n_rows,)](
+                out,
+                normed_x,
+                scaled_x,
+                inv_var,
+                x,
+                expanded_gamma,
+                expanded_beta,
+                x.stride(0),
+                expanded_gamma.stride(0),
+                expanded_beta.stride(0),
+                out.stride(0),
+                normed_x.stride(0),
+                scaled_x.stride(0),
+                inv_var.stride(0),
+                n_cols,
+                eps,
+                num_warps = num_warps,
+                BLOCK_SIZE = BLOCK_SIZE,
+            )
+            ctx.save_for_backward(scaled_x, normed_x, gamma, inv_var)
+        else:
+            layernorm_kernel_forward_inference[(n_rows,)](
+                out,
+                x,
+                expanded_gamma,
+                expanded_beta,
+                x.stride(0),
+                expanded_gamma.stride(0),
+                expanded_beta.stride(0),
+                out.stride(0),
+                n_cols,
+                eps,
+                num_warps = num_warps,
+                BLOCK_SIZE = BLOCK_SIZE,
+            )
 
-        ctx.save_for_backward(scaled_x, normed_x, gamma, inv_var)
         return out.view(*shape)
 
     @classmethod
@@ -454,11 +521,11 @@ class _layernorm(autograd.Function):
         )
 
         dx = dx.view(*shape)
-        return dx, dgamma, dbeta, None
+        return dx, dgamma, dbeta, None, None
 
-def layernorm(x, gamma, beta, eps = 1e-5, use_triton = False):
+def layernorm(x, gamma, beta, eps = 1e-5, use_triton = False, training = False):
     if use_triton:
-        out = _layernorm.apply(x, gamma, beta, eps)
+        out = _layernorm.apply(x, gamma, beta, eps, training)
     else:
         out = F.layer_norm(x, (x.shape[-1],), gamma, beta, eps = eps)
     return out
@@ -487,7 +554,7 @@ class Attention(nn.Module):
     def forward(self, x, mask = None, use_triton = None):
         use_triton = default(use_triton, self.use_triton)
         h = self.heads
-        x = layernorm(x, self.norm.weight, self.norm.bias, use_triton = use_triton)
+        x = layernorm(x, self.norm.weight, self.norm.bias, use_triton = use_triton, training = self.training)
 
         q, k, v = self.to_qkv(x).chunk(3, dim = -1)
         q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> (b h) n d', h = h), (q, k, v))
@@ -523,7 +590,7 @@ class FeedForward(nn.Module):
     def forward(self, x, use_triton = None):
         use_triton = default(use_triton, self.use_triton)
 
-        x = layernorm(x, self.norm.weight, self.norm.bias, use_triton = False)
+        x = layernorm(x, self.norm.weight, self.norm.bias, use_triton = False, training = self.training)
 
         x = fused_relu_squared(x, self.proj_in_weight, use_triton = use_triton)
         x = self.proj_out(x)
@@ -598,7 +665,7 @@ class Transformer(nn.Module):
             x = attn(x, mask = mask, use_triton = use_triton) + x
             x = ff(x, use_triton = use_triton) + x
 
-        x = layernorm(x, self.norm.weight, self.norm.bias, use_triton = use_triton)
+        x = layernorm(x, self.norm.weight, self.norm.bias, use_triton = use_triton, training = self.training)
         logits = self.to_logits(x)
 
         if not exists(labels):
